/////////////////////////////////////////////////////////////
// cacheb.cpp: Cache base class methods.
// Contains code not dependent of data type.
// Copyright(c) 1993 Azarona Software. All rights reserved.
/////////////////////////////////////////////////////////////
#include "cacheb.h"
#include "bucketb.h"

void Cacheb::Setup(Bucketb *b, int n, unsigned bkt_size)
// Using the n buckets pointed to by b, which are allocated
// and constructed, this routine finishes the construction
// by building a doubly-linked circular list out of them,
// and then configuring for an empty cache.
{
  Bucketb *p = b;
  for (int i = 0; i<n; i++) {
      p->MakeNull();
      p->prev = (Bucketb *)((char *)p - bkt_size);
      p->next = (Bucketb *)((char *)p + bkt_size);
      p = p->next;
  }
  // Make the list circular
  Bucketb *tail = (Bucketb *)((char *)b + (n-1)*bkt_size);
  tail->next = b;
  b->prev = tail;
  // Set up other cache variables
  nbuckets = n; head = b; 
  hits = 0; misses = 0; fast_binds = 0;
}

Cacheb::Cacheb(Bucketb *b, int n, unsigned bkt_size)
// Using the n buckets pointed to by b, which are allocated
// and constructed, this routine finishes the construction
// by building a doubly-linked circular list out of them,
// and then configuring for an empty cache.
: fptr(0)
{
  Setup(b, n, bkt_size);
}

void Cacheb::Connect(FmgrPtr &f)
// Connects cache to file f, assumed already opened.
{
  Clear(); // Flush any pending data
  fptr = f;
}

void Cacheb::Disconnect()
// Disconnects cache from the file it was connected to.
{
  if (fptr) Clear();  // Flush any pending data
  fptr = 0; // Connect to null object
}  

void Cacheb::MoveToFront(Bucketb *b)
// Logically move bucket to front of list, so that it is
// treated as the most recently reserved bucket.
// Shortcuts are used when possible.
{
  if (b == head->prev) { // b is tail
     head = head->prev;
  }
  else if (b != head) {
    // Unhook bucket from where it is
    b->prev->next = b->next;
    b->next->prev = b->prev;
    // Put it before head
    b->next = head;
    b->prev = head->prev;
    b->prev->next = b;
    head->prev = b;
    head = b;
  }
}

void Cacheb::Flush(int empty_bkts)
// Flushes all buckets in the cache. Makes them empty if
// empty_bkts is true. Checks for dangling coptrs.
{
  // Flush, starting at the front
  Bucketb *b = head;
  do {
    if (empty_bkts && b->IsLocked()) excp->THROW(DANGLINGPTR);
    if (fptr->ReadyForWriting()) b->Flush(*fptr);
    if (empty_bkts) b->MakeNull();
    b = b->next;
  } while(b != head);
}


void Cacheb::Clear()
// Flush all buckets in the cache, and then null them out.
// An exception is thrown if any dangling coptrs are found.
{
  if (fptr) Flush(1);
  hits = 0; misses = 0; fast_binds = 0;
}

Bucketb *Cacheb::AcquireBkt()
// Finds least recently reserved bucket that isn't locked,
// flushes the bucket, and moves it to the front. Passes
// back pointer to bucket or returns 0 if no bucket available.
{
  Bucketb *b = head->prev; // Least-recently reserved bucket
  while(1) {
    if (!b->IsLocked()) break;
    b = b->prev;
    if (b == head->prev) {
       excp->THROW(CACHEFULL); // Most likely will exit program
       return 0;               // Make compiler happy
    }
  }
  // Flush any data that might be in the bucket and move it to
  // the front so it becomes the most-recently reserved bucket.
  b->Flush(*fptr);
  MoveToFront(b);
  return b;
}

Bucketb *Cacheb::FindBkt(long addr)
// Searches for a bucket in the cache connected to file address
// addr. Returns pointer or 0 if no such bucket found.
{
  // Start search from front, (most recently reserved bucket)
  Bucketb *b = head;
  do {
    if (b->addr == addr) return b;
    b = b->next;
  } while(b != head);
  return 0;
}

Bucketb *Cacheb::ReserveBkt(long addr, int ensure_loaded)
// Reserves a bucket for file address addr. If
// ensure_loaded == 1, then if the data at addr is not
// already in bucket, it is loaded in. Moves bucket to the
// front. Passes back pointer to bucket, which may be 0
// if there is an error.
{
  Bucketb *b;

  if (addr == 0) return 0; // No sense reserving bucket

  b = FindBkt(addr);
  if (b == 0) { // Bkt data not loaded, so acquire a bucket
     misses++;
     b = AcquireBkt();
     if (b) {
        b->addr = addr;
        if (ensure_loaded && addr) {
           b->Fetch(*fptr);
        }
        // If failure fetching data, return null bucket to
        // signal failure, else lock the bucket.
        // Note: Any exception will probably cause us to
        // exit the program.
        if (!fptr->OK()) b = 0; else b->Lock();
     }
  }
  else {
    hits++;
    MoveToFront(b);
    b->Lock();
  }

  return b;
}

void Report(Cacheb &c, int full_report)
// Friend function that reports on the status of the cache.
{
  Bucketb *b = c.head;
  int cnt = 0, i = 0;
  do {
    if (b->IsLocked()) cnt++;
    if (full_report) {
       printf("Bucket[%02d]: Addr = %ld Refs = %d Dirty = %d\n",
              i, b->addr, b->refcnt, b->dirty);
    }
    i++; b = b->next;
  } while(b != c.head);

  long reservations = c.hits + c.fast_binds + c.misses;

  float ratio = 0.0;
  if (c.hits+c.fast_binds)
     ratio = (float)(c.hits+c.fast_binds) / (float)(reservations) * 100.0;
  printf("Fast binds/hits/misses: %ld / %ld / %ld (%6.2f%)\n",
         c.fast_binds, c.hits, c.misses, ratio);
  printf("Reservations:           %ld\n", reservations);
  printf("Buckets in use:         %d / %d\n", cnt, c.nbuckets);
}
